/* TABLIX, PGA highschool timetable generator                              */
/* Copyright (C) 2002 Tomaz Solc                                           */

/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License, or       */
/* (at your option) any later version.                                     */

/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */

/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

/* $Id: data.c,v 1.3.4.1 2003/06/08 22:40:20 avian Exp $ */

#include <stdlib.h>
#include <string.h>

#include "data.h"

int *cextbuff;                 /* Output buffers                           */
int *cnatbuff;
int *textbuff;

static int *conf_lookup;       /* Conflict lookup table and pointers to it */
static int **conf_lead;
static int **conf_start;

tuple **tuplemap;
int tuplenum=0;
int tuplemut=0;

teacher **tmap;
int tmapnum;

subject **smap;
int smapnum;

group **cmap;
int cmapnum;

classroom **rmap;
int rmapnum;

int days;
int periods;
int weight[WEIGHTMAX];

char *school_name;
char *school_address;
char *author;

/* The following functions arrange a chromosome along a timeline */
/* (tuplenum == -1) means free time                              */

/* cextbuff[cid*TIMES+time]=tuplenum       */
void extend_chromo(chromo *t)
{
        int b, c, cid;
        int *pnt;

        for(b=0;b<cmapnum*TIMES;b++) cextbuff[b]=-1;

        for(b=0;b<tuplenum;b++) {
                cid=tuplemap[b]->cid;

                for(pnt=conf_start[cid];pnt<conf_lead[cid];pnt++)
                cextbuff[(*pnt)*TIMES+t->inf[b].time]=b;
        }
}

/* textbuff[tid*TIMES+time]=tuplenum */
void textend_chromo(chromo *t)
{
        int b;

        for(b=0;b<tmapnum*TIMES;b++) textbuff[b]=-1;

        for(b=0;b<tuplenum;b++) {
                textbuff[tuplemap[b]->tid*TIMES+t->inf[b].time]=b;
        }
}

void oextend_chromo(chromo *t)
{
        int cid,time;
        int b,c;

        for(b=0;b<cmapnum*TIMES;b++) cextbuff[b]=-1;

        for(b=0;b<tuplenum;b++) {
                cid=tuplemap[b]->cid;
                time=t->inf[b].time;

                for(c=0;c<cmapnum;c++)  if (cmap[cid]->conflict[c]) {
                        cextbuff[c*TIMES+time]=b;
                        cnatbuff[c*TIMES+time]=c==cid;
                }
        }
}
/*
void oextend_chromo(chromo *t, int *echromo, int *native, int cid)
{
        int b;

        for(b=0;b<cmapnum*TIMES;b++) cextbuff[b]=-1;

        for(b=0;b<tuplenum;b++)
        if (cmap[cid]->conflict[tuplemap[b]->cid]) {
                echromo[t->inf[b].time]=b;
                native[t->inf[b].time]=tuplemap[b]->cid==cid;
        }
}
*/

/* *** TUPLE HANDLING FUNCTIONS *** */

void free_tuple()
{
        int c;

        for(c=0;c<tuplenum;c++) free(tuplemap[c]);
        free(tuplemap);
}

int add_tuple(int tid, int cid, int sid, int uid)
{
        int c;

        if (tuplenum==0) tuplemap=malloc(sizeof(*tuplemap)*TUPLEMAX);

        for(c=tuplenum;c>tuplemut;c--) tuplemap[c]=tuplemap[c-1];

        tuplemap[tuplemut]=malloc(sizeof(**tuplemap));

        tuplemap[tuplemut]->tid=tid;
        tuplemap[tuplemut]->cid=cid;
        tuplemap[tuplemut]->sid=sid;
        tuplemap[tuplemut]->imut_time=-uid;
        tuplemap[tuplemut]->imut_room=-1;

        tuplenum++;

        return(tuplemut++);
}

int add_tuple_imut(int tid, int cid, int sid, int time, int room)
{
        if (tuplenum==0) tuplemap=malloc(sizeof(*tuplemap)*TUPLEMAX);

        tuplemap[tuplenum]=malloc(sizeof(**tuplemap));

        tuplemap[tuplenum]->tid=tid;
        tuplemap[tuplenum]->cid=cid;
        tuplemap[tuplenum]->sid=sid;
        tuplemap[tuplenum]->imut_time=time;
        tuplemap[tuplenum]->imut_room=room;

        return(tuplenum++);
}

int find_tuple(int uid)
{
        static int tid=0;
        int b;

        b=tid;

        do {
                if (tuplemap[tid]->imut_time==(-uid)) return(tid);
                tid++;
                if (tid>=tuplemut) tid=0;
        } while (b!=tid);

        return(-1);
}

/* *** CLASSROOM MAP HANDLING FUNCTIONS *** */

int find_room(char *id)
{
        int c;

        for(c=0;c<rmapnum;c++) {
                if (!strcmp(rmap[c]->id, id)) return(c);
        }

        return(-1);
}

int has_cap(int rid, char *cap)
{
        int c;

        for(c=0;c<rmap[rid]->capnum;c++) {
                if (!strcmp(rmap[rid]->caplist[c], cap)) return(1);
        }

        return(0);
}

void free_building()
{
        int b,c;

        for(c=0;c<rmapnum;c++) {
                for(b=0;b<rmap[c]->capnum;b++) {
                        free(rmap[c]->caplist[b]);
                }
                free(rmap[c]->caplist);
                free(rmap[c]->id);
                free(rmap[c]);
        }

        free(rmap);
}

/* *** TEACHER MAP HANDLING FUNCTIONS *** */

int find_teacher(char *name)
{
        int c;

        for(c=0;c<tmapnum;c++) {
                if (!strcmp(tmap[c]->name, name)) return(c);
        }

        return(-1);
}

void free_teachers()
{
        int c;

        for(c=0;c<tmapnum;c++) {
                free(tmap[c]->name);
                free(tmap[c]);
        }

        free(tmap);
}

/* *** SUBJECT MAP HANDLING FUNCTIONS *** */

int find_subject(char *title, char *teacher)
{
        int c,tid;

        tid=find_teacher(teacher);

        for(c=0;c<smapnum;c++) {
                if (!strcmp(smap[c]->title, title)&&smap[c]->tid==tid) {
                        return(c);
                }
        }

        return(-1);
}

void free_subjects()
{
        int c;

        for(c=0;c<smapnum;c++) {
                free(smap[c]->title);
                free(smap[c]);
        }

        free(smap);
}

/* *** CLASS MAP HANDLING FUNCTIONS *** */

int find_class(char *name, int year)
{
        int c;

        for(c=0;c<cmapnum;c++) {
                if (!strcmp(cmap[c]->name, name)&&cmap[c]->year==year) {
                        return(c);
                }
        }

        return(-1);
}

/* must be run _after_ xmlsup_main() */
void data_init()
{
        int c, b;

        /* Output buffers */
        cextbuff=malloc(cmapnum*TIMES*sizeof(*cextbuff));
        cnatbuff=malloc(cmapnum*TIMES*sizeof(*cnatbuff));
        textbuff=malloc(tmapnum*TIMES*sizeof(*textbuff));

        /* Lookup tables */
        conf_lookup=malloc(cmapnum*cmapnum*sizeof(*conf_lookup));
        conf_lead=malloc(cmapnum*sizeof(*conf_lead));
        conf_start=malloc(cmapnum*sizeof(*conf_start));

        for(c=0;c<cmapnum;c++) {
                conf_start[c]=conf_lookup+c*cmapnum;
                conf_lead[c]=conf_start[c];
        }

        for(c=0;c<cmapnum;c++) {
                for(b=0;b<cmapnum;b++) {
                        if (cmap[c]->conflict[b]) {
                                *conf_lead[c]=b;
                                conf_lead[c]++;
                        }
                }
        }

}

void data_exit()
{
        free(cextbuff);
        free(cnatbuff);
        free(textbuff);

        free(conf_lookup);
        free(conf_lead);
        free(conf_start);
}
