/* TABLIX, PGA highschool timetable generator                              */
/* Copyright (C) 2002 Tomaz Solc                                           */

/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License, or       */
/* (at your option) any later version.                                     */

/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */

/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

/* $Id: master.c,v 1.13 2002/06/26 18:44:28 avian Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#ifdef HAVE_LIBPVM3
#include <pvm3.h>
#include <signal.h>
#endif

#include "main.h"

#define LINESPERDOT     10

char *cmd;
char prefix[256];

#ifdef HAVE_LIBPVM3
struct _sinfo {
        int send;
        int recv;
};

typedef struct _sinfo sinfo;

int nodenum=4;
int *child_tids;
sinfo *child_info;

struct pvmhostinfo *host_info;
int nhost;
int ctrlc;
#endif

void print_copyright()
{
        printf("\n");
        printf("This program is free software; you can redistribute it and/or modify\n");
        printf("it under the terms of the GNU General Public License as published by\n");
        printf("the Free Software Foundation; either version 2 of the License, or\n");
        printf("(at your option) any later version.\n");
        printf("\n");
        printf("This program is distributed in the hope that it will be useful,\n");
        printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
        printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
        printf("GNU General Public License for more details.\n");
        printf("\n");
        printf("You should have received a copy of the GNU General Public License\n");
        printf("along with this program; if not, write to the Free Software\n");
        printf("Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA\n");
        printf("\n");
        printf("$Id: master.c,v 1.13 2002/06/26 18:44:28 avian Exp $\n");
        printf("Compile time options:\n");
        #ifdef HAVE_LIBPVM3
        printf("        Parallel genetic algorithm (using PVM3)\n");
        #else
        printf("        Linear genetic algorith\n");
        #endif
        #ifdef HAVE_CONV
        printf("        Saving convergence info\n");
        #endif
        exit(0);
}

void print_syntax()
{
        printf("Usage: %s [OPTION]... [FILE]...\n", cmd);
        printf("\n");
        #ifdef HAVE_LIBPVM3
        printf("  -n NODES              set number of computing nodes (default 4)\n");
        #endif
        printf("  -o PREFIX             use PREFIX when naming output files\n");
        printf("  -r                    restore saved populations\n");
        printf("  -h                    display this help\n");
        printf("  -v                    display version, compile time options and\n");
        printf("                        copyright information\n");
        printf("\n");
        printf("Please report bugs to <tomaz.solc@siol.net>.\n");
        exit(0);
}

#ifdef HAVE_LIBPVM3
int child_from_tid(int tid)
{
        int c;

        for(c=0;c<nodenum;c++) {
                if(child_tids[c]==tid) return(c);
        }
        return(-1);
}

/* A child was killed. We have to tell its sibling   */
/* to send migration to another node. We also update */
/* child_info struct */
void child_killed(int num)
{
        int recv,send;

        send=child_info[num].send;
        recv=child_info[num].recv;

        child_info[send].recv=recv;
        child_info[recv].send=send;

        pvm_initsend(0);
        pvm_pkint(&child_tids[send], 1, 1);
        pvm_send(child_tids[recv], MSG_SIBLING);
}

void sighandler(int num)
{
        int c;

        if (ctrlc<1) {
                pvm_initsend(0);
                c=0;
                pvm_pkint(&c, 1, 1);
                pvm_mcast(child_tids, nodenum, MSG_SENDPOP);
                ctrlc++;
        } else {
                exit(1);
        }
}
#endif

int main(int argc, char *argv[])
{
        int c;
        #ifdef HAVE_LIBPVM3
        int d, a, msgtag, sender, sendernum;
        int tuplenum;
        int time, room;
        int rnum;

        double sum;
        int tospawn, running;
        int restore;

        FILE *xmlsrc;
        #ifdef HAVE_CONV
        FILE **convfiles;
        #endif
        char *buff;
        char fn[256];
        #endif

        cmd=argv[0];

        strcpy(prefix, "./");

        #ifdef HAVE_LIBPVM3
        restore=0;
        #endif

        printf("TABLIX version %s, PGA highschool timetable generator\n", VERSION);
        printf("Copyright (C) 2002 Tomaz Solc\n");

        while ((c=getopt(argc, argv, "hn:vo:r"))!=-1) {
                switch (c) {
                        case 'v': print_copyright();
                        case 'h':
                        case '?': print_syntax();
                                  exit(0);
                        case 'o': strncpy(prefix, optarg, 256);
                                  break;
                        #ifdef HAVE_LIBPVM3
                        case 'n': sscanf(optarg, "%d", &nodenum);
                                  break;
                        case 'r': restore=1;
                                  break;
                        #endif
                }
        }

        printf("\n");

        if (!(optind<argc)) {
                fprintf(stderr, "%s: Missing file name. Try '%s -h' for more information.\n", cmd, cmd);
                exit(1);
        }

        #ifndef HAVE_LIBPVM3
        execvp("tablix_slave", argv);

        perror(cmd);
        exit(1);
        #else

        child_tids=malloc(sizeof(*child_tids)*nodenum);
        child_info=malloc(sizeof(*child_info)*nodenum);
        if (child_tids==NULL||child_info==NULL) {
                perror(cmd);
                exit(1);
        }

        /* *** Smart spawning of slave nodes on different machines      *** */
        pvm_config(&nhost, NULL, &host_info);
        /* We get speed information from struct pvmhostinfo */

        sum=0;
        for(c=0;c<nhost;c++) sum+=host_info[c].hi_speed;

        running=0;
        for(a=0;a<nhost;a++) {
                /* Spawn slave nodes on one host*/
                tospawn=(int)(nodenum*host_info[a].hi_speed/sum+0.5);

                c=pvm_spawn("tablix_slave", NULL, PvmTaskHost, host_info[a].hi_name, tospawn, &child_tids[running]);
                if (c<tospawn) {
                        fprintf(stderr, "%s: %s: Some nodes failed to start. PVM error follows.\n", cmd, host_info[a].hi_name);
                }
                if (c<=0) pvm_perror(cmd); else running+=c;
        }
        if (running==0) {
                fprintf(stderr, "%s: All nodes failed.\n", cmd);
                pvm_exit();
                exit(1);
        }
        nodenum=running;

        printf("PGA using %d nodes on %d hosts\n", nodenum, nhost);
        printf("Multicasting XML data ");
        fflush(stdout);

        /* *** Send nodes XML data                                      *** */
        xmlsrc=fopen(argv[optind], "r");
        if (xmlsrc==NULL) {
                perror(cmd);
                for(c=0;c<nodenum;c++) pvm_kill(child_tids[c]);
                pvm_exit();
                exit(1);
        }

        buff=malloc(LINEBUFFSIZE);
        if (buff==NULL) {
                perror(cmd);
                fclose(xmlsrc);
                for(c=0;c<nodenum;c++) pvm_kill(child_tids[c]);
                pvm_exit();
                exit(1);
        }

        c=0;
        a=1;
        while (fgets(buff, LINEBUFFSIZE, xmlsrc)!=NULL) {
                pvm_initsend(0);
                pvm_pkint(&a, 1, 1);
                pvm_pkstr(buff);
                pvm_mcast(child_tids, nodenum, MSG_XMLDATA);
                c++;
                if (c%LINESPERDOT==0) {
                        printf(".");
                        fflush(stdout);
                }
        }

        pvm_initsend(0);
        a=0;
        strcpy(buff, "<!-- End of file -->\n");
        pvm_pkint(&a, 1, 1);
        pvm_pkstr(buff);
        pvm_mcast(child_tids, nodenum, MSG_XMLDATA);

        printf("+\n");

        fclose(xmlsrc);

        /* *** If we are restoring populations from a file              *** */
        if (restore) {
                printf("Restoring saved populations ");
                fflush(stdout);

                running=0;
                for(sendernum=0;sendernum<nodenum;sendernum++) {
                        pvm_initsend(0);

                        snprintf(fn, 256, "%ssave%d.txt", prefix, sendernum);
                        xmlsrc=fopen(fn, "r");

                        if (xmlsrc==NULL) {
                                tuplenum=-1;
                                pvm_pkint(&tuplenum, 1, 1);
                        } else {
                                fscanf(xmlsrc, "%d", &tuplenum);
                                pvm_pkint(&tuplenum, 1, 1);

                                fscanf(xmlsrc, "%d", &c);
                                pvm_pkint(&c, 1, 1);

                                for(c=0;c<POPSIZE;c++) {
                                        for(d=0;d<tuplenum;d++) {
                                                fscanf(xmlsrc, "%d %d", &time, &room);
                                                pvm_pkint(&time, 1, 1);
                                                pvm_pkint(&room, 1, 1);
                                        }
                                        fscanf(xmlsrc, "%s", fn);
                                        if (strcmp(fn, "+")) printf("XML input and saved population mismatch\n");
                                }
                                fclose(xmlsrc);
                                running++;
                        }
                        pvm_send(child_tids[sendernum], MSG_RESTOREPOP);
                }
                if (running<nodenum) {
                        printf("(%d nodes randomized)\n", nodenum-running);
                } else {
                        printf("\n");
                }
        } else {
                pvm_initsend(0);
                a=-1;
                pvm_pkint(&a, 1, 1);
                pvm_mcast(child_tids, nodenum, MSG_RESTOREPOP);
        }

        printf("Initializing nodes ( ");
        fflush(stdout);

        /* Send nodes sibling TID to send migration to */
        for(c=0;c<nodenum;c++) {
                pvm_initsend(0);
                if (c<(nodenum-1)) {
                        pvm_pkint(&child_tids[c+1], 1, 1);
                        child_info[c].send=c+1;
                } else {
                        pvm_pkint(&child_tids[0], 1, 1);
                        child_info[c].send=0;
                }
                pvm_send(child_tids[c], MSG_SIBLING);

                if (c>0) {
                        child_info[c].recv=c-1;
                } else {
                        child_info[c].recv=nodenum-1;
                }

                printf("%x ", child_tids[c]);
                fflush(stdout);
        }

        printf(")\n");
        printf("Parent ( %x ) listening...\n\n", pvm_mytid());

        #ifdef HAVE_CONV
        convfiles=malloc(nodenum*sizeof(*convfiles));
        if (convfiles==NULL) {
                perror(cmd);
                free(buff);
                /* We don't need to kill the nodes here. */
                /* They should detect this. */
                pvm_exit();
                exit(1);
        }

        for(c=0;c<nodenum;c++) {
                snprintf(fn, 256, "%sconv%d.txt", prefix, c);
                convfiles[c]=fopen(fn, restore?"a":"w");

                if (convfiles[c]==NULL) {
                        perror(cmd);
                        free(buff);
                        for(a=0;a<c;a++) fclose(convfiles[a]);
                        free(convfiles);
                        pvm_exit();
                        exit(1);
                }
        }
        #endif

        ctrlc=0;
        signal(SIGINT, sighandler);

        rnum=0;
        while (rnum<nodenum) {
                c=pvm_recv(-1, -1);
                pvm_bufinfo(c, NULL, &msgtag, &sender);

                sendernum=child_from_tid(sender);

                if (sendernum==-1) {
                        printf("*** [%x] received message from unknown node\n", sender);
                }

                if (msgtag==MSG_CHLDERR) {
                        pvm_upkstr(buff);
                        printf("*** [%x] reporting fatal error: %s\n", sender, buff);
                        rnum++;
                        child_killed(sendernum);
                } else
                if (msgtag==MSG_CHLDMSG) {
                        pvm_upkstr(buff);
                        printf("*** [%x] reporting: %s\n", sender, buff);
                } else
                if (msgtag==MSG_RESULTDATA) {
                        printf("[%x] uploading result ", sender);
                        fflush(stdout);
                        snprintf(fn, 256, "%sresult%d.xml", prefix, sendernum);

                        xmlsrc=fopen(fn, "w");
                        if (xmlsrc==NULL) {
                                perror(cmd);
                        }

                        a=0;
                        do {
                                pvm_upkint(&c, 1, 1);
                                pvm_upkstr(buff);

                                fprintf(xmlsrc, "%s", buff);

                                if (c) pvm_recv(sender, MSG_RESULTDATA);

                                a++;
                                if (a%LINESPERDOT==0) {
                                        printf(".");
                                        fflush(stdout);
                                }

                        } while (c);

                        fclose(xmlsrc);
                        printf("\n");

                        rnum++;

                        child_killed(sendernum);
                } else
                if (msgtag==MSG_INFO) {
                        pvm_upkint(&c, 1, 1);
                        pvm_upkint(&a, 1, 1);
                        pvm_upkint(&d, 1, 1);
                        printf("[%x] reports %d (%d) at %d %s\n", sender, c, d, a, (!d)?"(ok)":"");
                        #ifdef HAVE_CONV
                        fprintf(convfiles[sendernum], "%d\t%d\t%d\n", a, c, d);
                        fflush(convfiles[sendernum]);
                        #endif
                } else
                if (msgtag==MSG_POPDATA) {
                        printf("[%x] receiving population\n", sender);

                        snprintf(fn, 256, "%ssave%d.txt", prefix, sendernum);

                        xmlsrc=fopen(fn, "w");
                        if (xmlsrc==NULL) {
                                perror(cmd);
                        }

                        pvm_upkint(&tuplenum, 1, 1);
                        fprintf(xmlsrc, "%d\n", tuplenum);

                        pvm_upkint(&c, 1, 1);
                        fprintf(xmlsrc, "%d\n", c);

                        for(c=0;c<POPSIZE;c++) {
                                for(d=0;d<tuplenum;d++) {
                                        pvm_upkint(&time, 1, 1);
                                        fprintf(xmlsrc, "%d", time);
                                        pvm_upkint(&room, 1, 1);
                                        fprintf(xmlsrc, " %d\n", room);
                                }
                                fprintf(xmlsrc, "+\n");
                        }

                        fclose(xmlsrc);

                        rnum++;
                } else
                {
                        printf("*** [%x] Unknown message received. Perhaps a version mismatch?\n", sender);
                }
        }

        #ifdef HAVE_CONV
        for(c=0;c<nodenum;c++) fclose(convfiles[c]);
        free(convfiles);
        #endif

        free(child_tids);
        free(child_info);
        free(buff);

        printf("All the results were received.\n");

        pvm_exit();
        exit(0);
        #endif
}
