/* TABLIX, PGA highschool timetable generator                              */
/* Copyright (C) 2002 Tomaz Solc                                           */

/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License, or       */
/* (at your option) any later version.                                     */

/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */

/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

/* $Id: print.c,v 1.2 2002/06/22 13:23:11 avian Exp $ */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#undef HAVE_LIBPVM3

#include "data.h"
#include "error.h"
#include "xmlsup.h"
#include "output.h"

char *cmd;
chromo *parentpop;

void print_copyright()
{
        printf("\n");
        printf("This program is free software; you can redistribute it and/or modify\n");
        printf("it under the terms of the GNU General Public License as published by\n");
        printf("the Free Software Foundation; either version 2 of the License, or\n");
        printf("(at your option) any later version.\n");
        printf("\n");
        printf("This program is distributed in the hope that it will be useful,\n");
        printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
        printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
        printf("GNU General Public License for more details.\n");
        printf("\n");
        printf("You should have received a copy of the GNU General Public License\n");
        printf("along with this program; if not, write to the Free Software\n");
        printf("Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA\n");
        printf("\n");
        printf("$Id: print.c,v 1.2 2002/06/22 13:23:11 avian Exp $\n");
        exit(0);
}

void print_syntax()
{
        printf("Usage: %s [OPTION]... [FORMAT] [FILE]...\n", cmd);
        printf("\n");
        printf("  -o FILE               write output to file FILE instead of stdout\n");
        printf("  -h                    display this help\n");
        printf("  -v                    display version and copyright information\n");
        printf("\n");
        printf("Known output formats:\n");
        printf("  html                  default HTML\n");
        printf("\n");
        printf("Please report bugs to <tomaz.solc@siol.net>.\n");
        exit(0);
}

int main(int argc, char *argv[])
{
        int c;

        char xmlsrc[256];
        char output[256];

        FILE *out;

        strcpy(output, "-");

        cmd=argv[0];

        printf("TABLIX version %s, PGA highschool timetable generator\n", VERSION);
        printf("Copyright (C) 2002 Tomaz Solc\n");

        /* *** Parse command line options               *** */

        while ((c=getopt(argc, argv, "o:r"))!=-1) {
                switch (c) {
                        case 'v': print_copyright();
                        case 'h':
                        case '?': print_syntax();
                                  exit(0);
                        case 'o': strncpy(output, optarg, 256);
                                  break;
                }
        }

        if (!(optind<argc)) {
                fprintf(stderr, "%s: Missing output format. Try '%s -h' for more information.\n", cmd, cmd);
                exit(1);
        }
        if (!(optind<argc-1)) {
                fprintf(stderr, "%s: Missing file name. Try '%s -h' for more information.\n", cmd, cmd);
                exit(1);
        }

        /* *** Get XML input                            *** */
        strncpy(xmlsrc, argv[optind+1], 256);

        xmlsup_main(xmlsrc);

        /* *** Population                               *** */
        parentpop=malloc(sizeof(*parentpop));
        parentpop->inf=malloc(sizeof(*(parentpop->inf))*tuplenum);

        /* *** Buffers for extend functions             *** */
        cextbuff=malloc(cmapnum*TIMES*sizeof(*cextbuff));
        cnatbuff=malloc(cmapnum*TIMES*sizeof(*cnatbuff));
        textbuff=malloc(tmapnum*TIMES*sizeof(*textbuff));

        xmlsup_getchromo(parentpop);

        if (strcmp("-", output)) out=fopen(output, "w"); else out=stdout;

        if (out==NULL) {
                fprintf(stderr, "%s: Can't open output file.\n", cmd);
                exit(1);
        }

        if (!strcmp(argv[optind], "html")) {
                print_chromo_html(parentpop, out);
        } else {
                fprintf(stderr, "%s: Unknown output format. Try '%s -h' for more information.\n", cmd, cmd);
        }

        if (out!=stdout) fclose(out);

        free(cextbuff);
        free(cnatbuff);
        free(textbuff);

        xmlsup_exit();

        free(parentpop->inf);
        free(parentpop);

        return(0);
}
