/* TABLIX, PGA highschool timetable generator                              */
/* Copyright (C) 2002 Tomaz Solc                                           */

/* This program is free software; you can redistribute it and/or modify    */
/* it under the terms of the GNU General Public License as published by    */
/* the Free Software Foundation; either version 2 of the License, or       */
/* (at your option) any later version.                                     */

/* This program is distributed in the hope that it will be useful,         */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of          */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the           */
/* GNU General Public License for more details.                            */

/* You should have received a copy of the GNU General Public License       */
/* along with this program; if not, write to the Free Software             */
/* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA */

/* $Id: xmlsup.c,v 1.9.4.1 2003/06/08 22:40:20 avian Exp $ */

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xmlerror.h>

#include "data.h"
#include "xmlsup.h"
#include "error.h"

#if LIBXML_VERSION > 20411
  #define FAIL(msg,node) fatalxml(msg, xmlGetLineNo(node))
#else
  #define FAIL(msg,node) fatal(msg)
#endif

int have_uids;
int uid;

xmlDocPtr config;

/* *** INFO PARSER *** */

void parse_info(xmlNodePtr cur)
{
        while (cur!=NULL) {
                if (!strcmp(cur->name, "title")) {
                        school_name=xmlNodeGetContent(cur);
                } else if (!strcmp(cur->name, "address")) {
                        school_address=xmlNodeGetContent(cur);
                } else if (!strcmp(cur->name, "author")) {
                        author=xmlNodeGetContent(cur);
                }
                cur=cur->next;
        }
}

/* *** GENETIC PARAMETERS PARSER *** */

void parse_props(xmlNodePtr cur)
{
        int c, b;
        char *temp;

        while (cur!=NULL) {
                if (!strcmp(cur->name, "weight")) {
                        temp=xmlGetProp(cur, "id");
                        if (temp==NULL) FAIL("<weight> with no id", cur);

                        b=sscanf(temp, "%d", &c);
                        if (b<1) FAIL("invalid weight id", cur);
                        free(temp);

                        c--;
                        if (c>=WEIGHTMAX||c<0) FAIL("invalid weight id", cur);

                        temp=xmlGetProp(cur, "value");
                        if (temp==NULL) FAIL("<weight> with no value", cur);

                        b=sscanf(temp, "%d", &weight[c]);
                        if (b<1) FAIL("invalid <weight> value", cur);

                        free(temp);
                } else if (!strcmp(cur->name, "time")) {
                        temp=xmlGetProp(cur, "periods");
                        if (temp==NULL) FAIL("<time> with no periods", cur);

                        b=sscanf(temp, "%d", &periods);
                        if (b<1) FAIL("invalid periods in <time>", cur);
                        free(temp);

                        temp=xmlGetProp(cur, "days");
                        if (temp==NULL) FAIL("<time> with no days", cur);
                        b=sscanf(temp, "%d", &days);
                        if (b<1) FAIL("invalid days in <time>", cur);
                        free(temp);
                }
                cur=cur->next;
        }
}

/* *** BUILDING PARSER *** */

void parse_properties(xmlNodePtr cur)
{
        while (cur!=NULL) {
                if (!strcmp(cur->name, "capability")) {
                        rmap[rmapnum]->caplist[rmap[rmapnum]->capnum]=xmlNodeGetContent(cur);
                        rmap[rmapnum]->capnum++;
                }
                cur=cur->next;
        }
}

void parse_building(xmlNodePtr cur)
{
        int a,b,c,d;
        char *temp;

        rmap=malloc(sizeof(*rmap)*ROOMS);
        rmapnum=0;

        while (cur!=NULL) {
                if (!strcmp(cur->name, "classroom")) {
                        /* single classroom, can have capabilities */
                        rmap[rmapnum]=malloc(sizeof(*rmap[rmapnum]));
                        rmap[rmapnum]->caplist=malloc(sizeof(*rmap[rmapnum]->caplist)*MAXCAP);
                        rmap[rmapnum]->capnum=0;
                        rmap[rmapnum]->id=xmlGetProp(cur, "id");

                        if (rmap[rmapnum]->id==NULL) {
                                FAIL("<classroom> with no id", cur);
                        }

                        parse_properties(cur->children);

                        rmapnum++;
                } else if (!strcmp(cur->name, "classrooms")) {
                        /* multiple classrooms, no capabilities (yet) */
                        temp=xmlGetProp(cur, "start");
                        if (temp==NULL) {
                                FAIL("<classrooms> with no start", cur);
                        }
                        d=sscanf(temp, "%d", &a);
                        if (d<1) FAIL("invalid start in <classrooms>", cur);
                        free(temp);

                        temp=xmlGetProp(cur, "end");
                        if (temp==NULL) {
                                FAIL("<classrooms> with no end", cur);
                        }
                        d=sscanf(temp, "%d", &b);
                        if (d<1) FAIL("invalid end in <classrooms>", cur);
                        free(temp);

                        for(c=a;c<=b;c++) {
                                rmap[rmapnum]=malloc(sizeof(*rmap[rmapnum]));
                                rmap[rmapnum]->caplist=malloc(sizeof(*rmap[rmapnum]->caplist)*MAXCAP);
                                rmap[rmapnum]->capnum=0;
                                rmap[rmapnum]->id=malloc(10);
                                snprintf(rmap[rmapnum]->id, 10, "%d", c);

                                rmapnum++;
                        }

                }
                cur=cur->next;
        }
}

/* *** SUBJECTS & TEACHERS PARSER *** */

void parse_restrictions(xmlNodePtr cur)
{
        char *type, *temp;
        int c;

        smap[smapnum]->roomnum=0;
        smap[smapnum]->timenum=0;

        for(c=0;c<ROOMS;c++) smap[smapnum]->roomok[c]=1;
/*      for(c=0;c<TIMES;c++) smap[smapnum]->timeok[c]=1; this is not used */

        while (cur!=NULL) {
                if (!strcmp(cur->name, "restriction")) {
                        type=xmlGetProp(cur, "type");

                        if (type==NULL) {
                                FAIL("<restriction> with no type", cur);
                        }

                        temp=xmlNodeGetContent(cur);

                        if (!strcmp(type, "capability-place")) {
                                smap[smapnum]->roomnum=1;
                                for(c=0;c<rmapnum;c++) {
                                        /* room is OK only if it was OK   */
                                        /* previously and matches our new */
                                        /* criteria                       */

                                        smap[smapnum]->roomok[c]=smap[smapnum]->roomok[c]&&has_cap(c, temp);
                                }
                        }

                        free(temp);
                        free(type);
                }
                cur=cur->next;
        }
}

void parse_subjects(xmlNodePtr cur)
{
        char *temp;
        int c;

        tmap=malloc(sizeof(*tmap)*TIDS);
        tmapnum=0;

        smap=malloc(sizeof(*smap)*MAXSUBJECTS);
        smapnum=0;

        while (cur!=NULL) {
                if (!strcmp(cur->name, "subject")) {
                        smap[smapnum]=malloc(sizeof(**smap));
                        smap[smapnum]->title=xmlGetProp(cur, "title");

                        if (smap[smapnum]->title==NULL) {
                                FAIL("<subject> with no title", cur);
                        }

                        temp=xmlGetProp(cur, "teacher");

                        if (temp==NULL) {
                                FAIL("<subject> with no teacher", cur);
                        }

                        c=find_teacher(temp);
                        if (c==-1) {
                                /* this is a new teacher, create a new entry */
                                tmap[tmapnum]=malloc(sizeof(*tmap[tmapnum]));
                                tmap[tmapnum]->name=temp;
                                tmap[tmapnum]->perday=0;
                                smap[smapnum]->tid=tmapnum;
                                tmapnum++;
                        } else {
                                /* known teacher, assign old tid */
                                smap[smapnum]->tid=c;
                        }

                        parse_restrictions(cur->children);

                        smapnum++;
                }
                cur=cur->next;
        }
}

/* *** TIMETABLE & CLASSES PARSER  *** */

void parse_conflicts(xmlNodePtr cur)
{
        xmlNodePtr start;
        char *temp, *temp1, *name;
        int c,b;
        int curcid, confcid;

        curcid=-1;

        start=cur;

        /* First we need to find a cid for this class */
        while (cur!=NULL) {
                if (!strcmp(cur->name, "class")) {
                        name=xmlGetProp(cur, "name");
                        temp=xmlGetProp(cur, "year");
                        b=sscanf(temp, "%d", &c);
                        if (b<1) FAIL("invalid year in <class>", cur);
                        curcid=find_class(name, c);
                        free(name);
                        free(temp);
                }
                cur=cur->next;
        }

        if (curcid==-1) {
                FAIL("<timetable> without <class>", start);
        }

        /* Only then can we start all over again and */
        /* realy parse those conflicts */
        cur=start;
        while (cur!=NULL) {
                if (!strcmp(cur->name, "restriction")) {
                        temp=xmlGetProp(cur, "type");
                        if (temp==NULL) FAIL("<restriction> with no type", cur);

                        if (!strcmp(temp, "conflicts-with")) {
                                name=xmlGetProp(cur, "name");
                                temp1=xmlGetProp(cur, "year");
                                b=sscanf(temp1, "%d", &c);
                                if (b<1) FAIL("invalid year in <restriction>", cur);
                                confcid=find_class(name, c);

                                if (confcid==-1) {
                                        FAIL("invalid conflicts-with restriction", cur);
                                }

                                free(temp1);
                                free(name);

                                cmap[curcid]->conflict[confcid]=1;
                                cmap[confcid]->conflict[curcid]=1;
                        }
                        free(temp);
                }
                cur=cur->next;
        }
}

void parse_timetable(xmlNodePtr cur)
{
        group *curclass;
        char *temp, *temp1;
        int sid,c,day,period,room,b;

        curclass=malloc(sizeof(*curclass));
        cmap[cmapnum]=curclass;

        for(c=0;c<CIDS;c++) curclass->conflict[c]=0;
        curclass->conflict[cmapnum]=1;

        curclass->perday=0;

        while (cur!=NULL) {
                if (!strcmp(cur->name, "class")) {
                        curclass->name=xmlGetProp(cur, "name");

                        if (curclass->name==NULL) {
                                FAIL("<class> with no name", cur);
                        }

                        temp=xmlGetProp(cur, "year");
                        b=sscanf(temp, "%d", &curclass->year);
                        if (b<1) FAIL("invalid year in <classs>", cur);
                        free(temp);

                } else if (!strcmp(cur->name, "subject")) {
                        temp=xmlGetProp(cur, "title");
                        temp1=xmlGetProp(cur, "teacher");

                        sid=find_subject(temp, temp1);

                        if (sid==-1) {
                                FAIL("<subject> in <timetable> without definition", cur);
                        }

                        free(temp);
                        free(temp1);

                        if (have_uids) {
                                temp=xmlGetProp(cur, "uid");
                                b=sscanf(temp, "%d", &uid);
                                if (b<1) FAIL("Invalid uid", cur);
                                free(temp);
                        }

                        temp=xmlGetProp(cur, "perweek");

                        if (temp==NULL) {
                                temp1=xmlGetProp(cur, "period");
                                if (temp1==NULL) FAIL("<subject> without perweek or period", cur);
                                b=sscanf(temp1, "%d", &period);
                                if (b<1) FAIL("invalid period in <subject>", cur);
                                free(temp1);

                                temp1=xmlGetProp(cur, "day");
                                if (temp1==NULL) FAIL("<subject> without perweek or day", cur);
                                b=sscanf(temp1, "%d", &day);
                                if (b<1) FAIL("invalid day in <subject>", cur);
                                free(temp1);

                                temp1=xmlGetProp(cur, "room");
                                if (temp1==NULL) FAIL("<subject> without perweek or room", cur);
                                room=find_room(temp1);
                                if (room==-1) FAIL("<subject> with invalid room", cur);
                                free(temp1);

                                add_tuple_imut(smap[sid]->tid, cmapnum, sid, PERIODS*day+period, room);
                                curclass->perday++;
                                tmap[smap[sid]->tid]->perday++;
                        } else {
                                b=sscanf(temp, "%d", &c);
                                if (b<1) FAIL("invalid perweek in <subject>", cur);

                                free(temp);

                                if (!have_uids) {
                                        temp=malloc(256*sizeof(*temp));
                                        snprintf(temp, 256, "%d", uid);
                                        xmlNewProp(cur, "uid", temp);
                                        free(temp);
                                }

                                for(;c>0;c--) {
                                        add_tuple(smap[sid]->tid, cmapnum, sid, uid++);
                                        curclass->perday++;
                                        tmap[smap[sid]->tid]->perday++;
                                }
                        }
                }
                cur=cur->next;
        }

        curclass->perday/=DAYS;

        cmapnum++;
}

/* *** Error reporting handler for XML parser *** */

void myhandler(void *ctx, const char *msg, ...) {
    va_list args;
    char pmsg[256];

    va_start(args, msg);
    vsnprintf(pmsg, 256, msg, args);
    va_end(args);

    notify(pmsg);
}


/* *** MAIN *** */

void xmlsup_main(char *what)
{
        xmlNodePtr cur, building, subjects, population, props, info;

        int c;

        xmlLineNumbersDefault(1);
        xmlSetGenericErrorFunc(NULL, myhandler);

        config=xmlParseFile(what);
        if (config==NULL) fatal("Parse error");

        cur=xmlDocGetRootElement(config);

        if (strcmp(cur->name, "school")) {
                FAIL("Root element mismatch", cur);
        }

        cur=cur->children;

        building=NULL;
        subjects=NULL;
        population=NULL;
        props=NULL;
        info=NULL;

        while (cur!=NULL) {
                if (!strcmp(cur->name, "building")) {
                        building=cur->children;
                } else if (!strcmp(cur->name, "subjects")) {
                        subjects=cur->children;
                } else if (!strcmp(cur->name, "population")) {
                        population=cur->children;
                } else if (!strcmp(cur->name, "parameters")) {
                        props=cur->children;
                } else if (!strcmp(cur->name, "info")) {
                        info=cur->children;
                }

                cur=cur->next;
        }

        if (building==NULL) {
                fatal("no <building>");
        }
        if (subjects==NULL) {
                fatal("no <subjects>");
        }
        if (population==NULL) {
                have_uids=0;
                uid=1;
        } else {
                have_uids=1;
        }

        /* Default weights, parameters */
        days=5;
        periods=7;
        /* Two tuples in the same time and space */
        weight[0]=20;
        /* One teacher or one class should be in two */
        /*places in the same time */
        weight[1]=20;
        /* Add one weight for each free hour a class */
        /* has in the morning */
        weight[2]=1;
        /* Add one weight for each trek from one room */
        /* to another */
        weight[3]=1;
        /* Add one weight for each hole in the timetable */
        weight[4]=8;
        /* Add one weight for each unsatisfied restriction */
        weight[5]=30;
        /* Add one weight for each hour too much/missing in a day */
        weight[8]=1;
        /* Teacher's timetable */
        /* Add one weight for each free hour a teacher */
        /* has in the morning */
        weight[6]=0;
        /* Add one weight for each hole in the timetable */
        weight[7]=5;
        /* Add one weight for each hour too much/missing in a day */
        weight[9]=1;
        /* This is added to all impossible timetables */
        weight[10]=25;

        if (!(props==NULL)) {
                parse_props(props);
        }

        if (!(info==NULL)) {
                parse_info(info);
        } else {
                school_name=NULL;
                school_address=NULL;
                author=NULL;
        }

        parse_building(building);
        parse_subjects(subjects);

/*
        for(c=0;c<smapnum;c++) {
                printf("%s - %d\n", smap[c]->title, smap[c]->roomnum);
        }
        getchar();
*/

        /* begin parsing timetables */

        cmap=malloc(sizeof(*cmap)*CIDS);
        cmapnum=0;

        cur=xmlDocGetRootElement(config);
        cur=cur->children;

        while (cur!=NULL) {
                if (!strcmp(cur->name, "timetable")) {
                        parse_timetable(cur->children);
                }

                cur=cur->next;
        }

        /* begin parsing conflicts */

        cur=xmlDocGetRootElement(config);
        cur=cur->children;

        while (cur!=NULL) {
                if (!strcmp(cur->name, "timetable")) {
                        parse_conflicts(cur->children);
                }

                cur=cur->next;
        }

        for(c=0;c<tmapnum;c++) tmap[c]->perday/=DAYS;

        /* done parsing */

}

void xmlsup_getchromo(chromo *pop)
{
        xmlNodePtr cur, chromo;
        int period, day, uid, tid, b;
        char *room;
        char *temp;

        chromo=NULL;

        cur=xmlDocGetRootElement(config);
        cur=cur->children;

        while (cur!=NULL) {
                if (!strcmp(cur->name, "chromo")) {
                        chromo=cur;
                }
                cur=cur->next;
        }

        if (chromo==NULL) fatal("no <chromo>");

        temp=xmlGetProp(chromo, "grade");
        if (temp==NULL) FAIL("<chromo> without grade", chromo);
        b=sscanf(temp, "%d", &(pop->grade));
        if (b<1) FAIL("invalid grade in <chromo>", chromo);
        free(temp);

        cur=chromo->children;

        while (cur!=NULL) {
                if (!strcmp(cur->name, "gen")) {
                        temp=xmlGetProp(cur, "period");
                        if (temp==NULL) FAIL("<gen> without period", cur);
                        b=sscanf(temp, "%d", &period);
                        if (b<1) FAIL("invalid period in <gen>", cur);
                        free(temp);

                        temp=xmlGetProp(cur, "day");
                        if (temp==NULL) FAIL("<gen> without day", cur);
                        b=sscanf(temp, "%d", &day);
                        if (b<1) FAIL("invalid day in <gen>", cur);
                        free(temp);

                        room=xmlGetProp(cur, "room");
                        if (room==NULL) FAIL("<gen> without room", cur);

                        temp=xmlGetProp(cur, "uid");
                        if (temp==NULL) FAIL("<gen> without uid", cur);
                        b=sscanf(temp, "%d", &uid);
                        if (b<1) FAIL("invalid uid in <gen>", cur);
                        free(temp);

                        tid=find_tuple(uid);
                        if (tid==-1) FAIL("Invalid uid", cur);
                        pop->inf[tid].time=day*PERIODS+period;
                        pop->inf[tid].room=find_room(room);

                        free(room);
                }
                cur=cur->next;
        }
	
	/* bug fix by Jaume Obrador */
	for(b=tuplemut;b<tuplenum;b++) {
		pop->inf[b].time=tuplemap[b]->imut_time;
		pop->inf[b].room=tuplemap[b]->imut_room;
	}
}

void xmlsup_addchromo(chromo *pop)
{
        xmlNodePtr cur, chromo, gen;
        int c, b;
        char temp[256];

        cur=xmlDocGetRootElement(config);

        xmlNodeAddContent(cur, "\t");
        chromo=xmlNewChild(cur, NULL, "chromo", NULL);
        xmlNodeAddContent(cur, "\n");

        snprintf(temp, 256, "%d", pop->grade);
        xmlNewProp(chromo, "grade", temp);

        for(b=0;b<tuplemut;b++) {
                xmlNodeAddContent(chromo, "\n\t\t");
                gen=xmlNewChild(chromo, NULL, "gen", NULL);

                snprintf(temp, 256, "%d", pop->inf[b].time%PERIODS);
                xmlNewProp(gen, "period", temp);

                snprintf(temp, 256, "%d", pop->inf[b].time/PERIODS);
                xmlNewProp(gen, "day", temp);

                xmlNewProp(gen, "room", rmap[pop->inf[b].room]->id);

                snprintf(temp, 256, "%d", -tuplemap[b]->imut_time);
                xmlNewProp(gen, "uid", temp);
        }
        xmlNodeAddContent(chromo, "\n\t");
}

void xmlsup_dump(FILE *f)
{
        xmlDocDump(f, config);
}

void xmlsup_exit()
{
        xmlFreeDoc(config);

        free_teachers(); /* :) */
        free_subjects();
        free_building();
}
